################################################################
#
#   Assembly Line - Controls
#
################################################################

from pygame.display import get_surface
from albow.dialogs import Modal, Dialog
from albow.controls import Label, Button, Column, ValueView
from albow.palette_view import PaletteView
from albow.resource import get_font
from albow.utils import blit_in_rect
from game import game

#-----------------------------------------------------------------------

class Multichoice(PaletteView):

	highlight_style = 'fill'
	align = 'c'

	def __init__(self, cell_size, values, attribute, model, **kwds):
		PaletteView.__init__(self, cell_size, 1, len(values), **kwds)
		self.values = values
		self.attribute = attribute
		self.model = model

	def num_items(self):
		return len(self.values)
	
	def item_is_selected(self, n):
		return self.get_value() == self.values[n]
	
	def click_item(self, n, e):
		self.set_value(self.values[n])
	
	def get_value(self):
		return getattr(self.model, self.attribute)
	
	def set_value(self, x):
		setattr(self.model, self.attribute, x)

#-----------------------------------------------------------------------

class ImageMultichoice(Multichoice):

	def __init__(self, images, values, **kwds):
		image0 = images[0]
		w, h = image0.get_size()
		d = 2 * self.predict(kwds, 'margin')
		cell_size = w + d, h + d
		Multichoice.__init__(self, cell_size, values, **kwds)
		self.images = images

	def draw_item(self, surf, n, rect):
		image = self.images[n]
		#d = self.margin
		#surf.blit(image, rect.left + d, rect.top + d)
		blit_in_rect(surf, image, rect, self.align, self.margin)
	
#-----------------------------------------------------------------------

class TextMultichoice(Multichoice):


	def __init__(self, values, labels = None, **kwds):
		if not labels:
			labels = map(str, values)
		font = self.predict_font(kwds)
		widths = [w for (w, h) in map(font.size, labels)]
		d = 2 * self.predict(kwds, 'margin')
		cw = max(widths) + d
		ch = font.get_linesize() + d
		Multichoice.__init__(self, (cw, ch), values, **kwds)
		self.labels = labels

	def draw_item(self, surf, n, rect):
		buf = self.font.render(self.labels[n], True, self.fg_color)
		blit_in_rect(surf, buf, rect, self.align, self.margin)

#-----------------------------------------------------------------------

class MachineEditor(Dialog):

	title_font = get_font(18, "VeraBd.ttf")
	
	def __init__(self, machine):
		Dialog.__init__(self)
		self.machine = machine
		title = Label("Configure " + machine.title, font = self.title_font)
		inner = self.create_contents(machine)
		title_and_inner = Column([title, inner], spacing = 16, align = 'c')
		ok_button = Button("OK", self.ok)
		contents = Column([title_and_inner, ok_button], align = 'r')
		contents.topleft = (16, 16)
		self.add(contents)
		self.shrink_wrap()

	def ok(self):
		self.dismiss(True)

#-----------------------------------------------------------------------

class ProductKeyField(ValueView):

	border_width = 1
	margin = 3

	def __init__(self, width = 200, **kwds):
		ValueView.__init__(self, width = width, **kwds)
	
	def mouse_down(self, e):
		dlog = ProductKeyDialog(width = self.width,
			attribute = self.attribute, model = self.model,
			bg_color = self.inherited('bg_color'))
		p = self.local_to_global((0, 0))
		dlog.topleft = p
		sw, sh = get_surface().get_size()
		if dlog.bottom > sh:
			dlog.bottom = sh
		if dlog.top < 0:
			dlog.top = 0
		dlog.present(centered = False)
	
	def format_value(self, key):
		if key:
			return game.product_with_key(key).title
		else:
			return ""

#-----------------------------------------------------------------------

class ProductKeySelector(PaletteView):

	cell_width = 150
	lmargin = 5
	highlight_style = 'fill'
	fg_color = (0, 0, 0)
	bg_color = (255, 255, 255)
	sel_color = (62, 188, 254)
	border_width = 2
	attribute = None
	model = None

	def __init__(self, width, height = 300, **kwds):
		font = self.predict_font(kwds)
		cw = width
		ch = font.get_linesize() + 4
		nr = height // ch
		PaletteView.__init__(self, (cw, ch), nr, 1, scrolling = True, **kwds)
		self.scroll_to_value(self.get_value())
	
	def scroll_to_value(self, key):
		product = game.product_with_key(key)
		if product:
			n = game.products.index(product)
			self.scroll_to_item(n)
	
	def num_items(self):
		return len(game.products)
	
	def item_is_selected(self, n):
		return game.products[n].key == self.get_value()
	
	def click_item(self, n, e):
		self.set_value(game.products[n].key)
	
	def draw_item(self, surf, n, rect):
		product = game.products[n]
		buf = self.font.render(product.title, True, self.fg_color)
		r = buf.get_rect()
		r.midleft = rect.left + self.lmargin, rect.centery
		surf.blit(buf, r)
	
	def get_value(self):
		return getattr(self.model, self.attribute)
	
	def set_value(self, x):
		setattr(self.model, self.attribute, x)

#-----------------------------------------------------------------------

class ProductKeyDialog(Modal, ProductKeySelector):

	def click_item(self, n, e):
		ProductKeySelector.click_item(self, n, e)
		self.dismiss(True)
