################################################################
#
#   Assembly Line - Factory
#
################################################################

from pygame import Rect
from albow.resource import get_image
from transformation import Transformation, add, sub, neg, dot
from machine import MachineNotConfiguredError

class FactoryError(Exception):
	pass

#------------------------------------------------------------------------

class Cell(object):
	#  layers   [Machine]
	#  #parts    [Part]
	
	def __init__(self):
		self.layers = [None, None]
		self.parts = []

#	def __setstate__(self, d):
#		self.__dict__.update(d)
#		if 'parts' not in d:
#			self.parts = []

#------------------------------------------------------------------------

class Factory(object):
	#  grid_size               (int, int)
	#  cells                   list[cx][cy] of Cell
	#  machines                list[layer] of [Machine]
	#  parts                   [Part]
	#  frameno                 int
	#  running                 boolean
	#  error_position          (x, y)
	#  error_message           string
	#  inspected_machine       Machine
	
	inspection_badge = get_image("inspection_badge.png")

	cell_size = (32, 32)
	grid_size = (24, 18)
	frameno = 0
	running = False
	error_position = None
	error_message = ""
	inspected_machine = None
	
	def __init__(self):
		cw, ch = self.grid_size
		self.cells = [
			[Cell() for cy in xrange(ch)]
				for cx in xrange(cw)]
		self.machines = [[], []]
		self.parts = []
		#<<<
		#from machines.input_hopper import InputHopper
		#m = InputHopper()
		#t = Transformation(0, (32, 64))
		#self.add_machine(m, t)
		#>>>
	
	def __setstate__(self, d):
		self.__dict__.update(d)
		if 'parts' not in d:
			self.parts = []
	
	def add_part(self, part, position):
		part.position = position
		self.parts.append(part)
		#self.update_part_handler(part)
	
	def move_part(self, part, v):
		part.position = add(v, part.position)
	
	def move_part_to(self, part, p):
		part.position = p
	
	def remove_part(self, part):
		try:
			self.parts.remove(part)
		except ValueError:
			pass
		handler = part.handler
		if handler:
			handler.part_removed(part)
			part.handler = None
	
	#def update_part_handler(self, part):
	#	cell = self.cell_at(part.position)
	#	if cell:
	#		part.handler = cell.layers[0]

	def start(self):
		self.clear_error()
		self.check_ready()
		self.connect_conveyors()
		self.running = True
	
	def stop(self):
		self.running = False
	
	def check_ready(self):
		try:
			for machine in self.all_machines():
				machine.check_ready()
		except MachineNotConfiguredError:
			self.machine_error(machine, "%s is not fully configured." % machine.title)
	
	def find_machine_at(self, p, layer = None):
		cell = self.cell_at(p)
		if cell:
			if layer is not None:
				layers = [layer]
			else:
				layers = (0, 1)
			for i in layers:
				machine = cell.layers[i]
				if machine:
					return machine
			for machine in cell.layers:
				if machine:
					return machine
	
	def unpurchased_machines(self):
		result = []
		for machine in self.all_machines():
			if not machine.purchased:
				result.append(machine)
			machine.find_unpurchased_components(result)
		return result
	
	def all_machines(self):
		for layer in self.machines:
			for machine in layer:
				yield machine

	def add_machine(self, machine, t):
		lc = []
		okay, rects, t = self.new_machine_position(machine, t, lc)
		#print "Factory.add_machine: okay =", okay, "rects =", rects ###
		if okay:
			#print "Factory.add_machine: affected cells =", lc ###
			machine.set_transformation(t)
			for i in (0, 1):
				if machine.in_layer(i):
					self.machines[i].append(machine)
			#self.machines[machine.layer].append(machine)
			for i, cell in lc:
				cell.layers[i] = machine
			machine.update_neighbourhood(self)
		return okay
	
	def remove_machine(self, machine):
		#print "Factory.remove_machine:", machine ###
		machine.removed_from_factory()
		for layer in self.machines:
			try:
				layer.remove(machine)
				#print "...removed from main machine list" ###
			except ValueError:
				pass
		for row in self.cells:
			for cell in row:
				layers = cell.layers
				for i in (0, 1):
					if layers[i] is machine:
						#print "...removing from layer", i, "of", cell ###
						layers[i] = None
		for part in self.parts:
			if part.handler is machine:
				part.handler = None
		if machine is self.inspected_machine:
			self.inspected_machine = None

	def new_machine_position(self, machine, t, affected_cells = None):
		mtype = machine.__class__
		return self.new_machine_type_position(mtype, t, affected_cells, machine)

	def new_machine_type_position(self, machine_type, t,
			affected_cells = None, ignore_machine = None):
		bounds = machine_type.bounding_rect(t)
		pos = bounds.topleft
		snapped_pos = self.snap_to_grid(pos)
		if machine_type.on_left:
			snapped_pos = (0, snapped_pos[1])
		delta = sub(snapped_pos, pos)
		t = Transformation(t.rotation, add(delta, t.translation))
		rects = machine_type.covered_rects(t)
		okay = self.space_is_available_for_rects(rects, affected_cells, ignore_machine)
		return okay, rects, t
	
	def space_is_available_for_rects(self, rects, affected_cells = None, ignore_machine = None):
		factory_bounds = self.bounding_rect()
		for i, layer_rects in enumerate(rects):
			for r in layer_rects:
				#print "Factory.space_available_for_rects: checking", r, "in", factory_bounds ###
				if not factory_bounds.contains(r):
					return False
				for cell in self.cells_covered_by_rect(r):
					#print "Factory.space_available_for_rects: cell.layers =", cell.layers ###
					m = cell.layers[i]
					if m and m is not ignore_machine:
						return False
					if affected_cells is not None:
						affected_cells.append((i, cell))
		return True

	def cells_covered_by_rect(self, r):
		cl, ct = self.pixel_to_cell_coords(r.topleft)
		cw, ch = self.pixel_to_cell_coords(r.size)
		#print "Factory.cells_covered_by_rect: cl =", cl, "ct =", ct, "cw =", cw, "ch =", ch ###
		cells = self.cells
		result = [cells[cx][cy]
			for cx in xrange(cl, cl + cw)
				for cy in xrange(ct, ct + ch)]
		#print "...", result ###
		return result
	
	def snap_to_grid(self, (x, y)):
		wc, hc = self.cell_size
		sx = ((x + wc // 2) // wc) * wc
		sy = ((y + hc // 2) // hc) * hc
		#print "Factory.snap_to_grid:", (x, y), "-->", (sx, sy) ###
		return sx, sy
	
	def pixel_to_cell_coords(self, (x, y)):
		cw, ch = self.cell_size
		return x // cw, y // ch
	
	def cell_at(self, p):
		cw, ch = self.grid_size
		cx, cy = self.pixel_to_cell_coords(p)
		if 0 <= cx < cw and 0 <= cy < ch:
			return self.cells[cx][cy]
	
	def machine_at(self, p, layer_no):
		cell = self.cell_at(p)
		if cell:
			return cell.layers[layer_no]
	
	def bounding_rect(self):
		cw, ch = self.grid_size
		wc, hc = self.cell_size
		return Rect(0, 0, cw * wc, ch * hc)
	
	def draw(self, surf, t):
		layers = self.machines
		self.draw_machines(surf, layers[0], t, 0)
		self.draw_parts(surf, t)
		self.draw_machines(surf, layers[1], t, 1)
		self.draw_inspection_badge(surf, t)
		
	def draw_machines(self, surf, machines, t, layer):
		for machine in machines:
			#print "Drawing", machine ###
			machine.draw(surf, t, layer)
	
	def draw_inspection_badge(self, surf, t):
		machine = self.inspected_machine
		if machine:
			r = machine.machine_bounding_rect(t)
			r.move_ip(4, 4)
			surf.blit(self.inspection_badge, r)
	
	def draw_parts(self, surf, t):
		v = t.translation
		for part in self.parts:
			p = add(v, part.position)
			pixels = part.pixels
			r = pixels.get_rect()
			r.center = p
			surf.blit(part.pixels, r)
	
	def begin_frame(self):
		if self.running:
			#print "Factory: begin_frame" ###
			for machine in self.all_machines():
				machine.begin_frame(self)
			for part in self.parts:
				if not part.handler:
					self.handle_part(part)

	def handle_part(self, part):
		#print "Factory: Handling part", part, "with handler", part.handler ###
		#machine = part.handler
		#if machine:
		#	machine.handle_part(self, part)
		#else:
			cell = self.cell_at(part.position)
			if cell:
				layers = cell.layers
				for i in 1, 0:
					machine = layers[i]
					if machine:
						#if machine.handle_part(self, part):
						#	break
						machine.handle_part(self, part)
						if part.handler:
							break
	
	def cell(self, (cx, cy)):
		return self.cells[cx][cy]
	
	def cell_coords_in_bounds(self, (cx, cy)):
		cw, ch = self.grid_size
		return 0 <= cx < cw and 0 <= cy < ch
	
	def origin_cell_coords(self, machine):
		return self.pixel_to_cell_coords(machine.translation)
	
	def cell_center_coords(self, (cx, cy)):
		wc, hc = self.cell_size
		return cx * wc + wc // 2, cy * hc + hc // 2
	
	def connect_conveyors(self):
		#print "Factory.connect_conveyors" ###
		for machine in self.machines[0]:
			if machine.is_conveyor:
				machine.drive = None
		for drive in self.machines[0]:
			if drive.is_conveyor_drive:
				#print "Found conveyor drive", drive ###
				cxy = self.origin_cell_coords(drive)
				dxy = drive.adjacent_direction()
				cxy = add(cxy, dxy)
				belt = self.cell(cxy).layers[0]
				#print "Drive origin in cell", cxy, "adjacent", dxy, "belt", belt ###
				if belt and belt.is_conveyor_end:
					fxy = belt.forward_direction()
					#print "Forward direction", fxy ###
					if dot(fxy, dxy) == 0:
						#print "Belt is being driven" ###
						belt.drive = drive
						if belt.which_end == 0:
							self.connect_conveyor_to_drive(cxy, fxy, drive)
						else:
							self.connect_conveyor_to_drive(cxy, neg(fxy), drive)
						
	def connect_conveyor_to_drive(self, cxy, fxy, drive):
		while 1:
			cxy1 = add(cxy, fxy)
			if not self.cell_coords_in_bounds(cxy1):
				break
			cxy = cxy1
			belt = self.cell(cxy).layers[0]
			if belt and belt.is_conveyor:
				#print "Factory: connecting belt", belt, "to drive", drive ###
				belt.drive = drive
			else:
				break
			#print "Is conveyor end =", belt.is_conveyor_end ###
			if belt.is_conveyor_end:
				return
		self.cell_error(cxy, "Missing conveyor belt end section")
	
	def cell_error(self, cxy, mess):
		self.error_position = self.cell_center_coords(cxy)
		self.error_message = mess
		raise FactoryError(mess)
	
	def machine_error(self, machine, mess):
		p = machine.global_origin()
		self.factory_error(p, mess)
		
	def factory_error(self, p, mess):
		self.error_position = p
		self.error_message = mess
		raise FactoryError(mess)
	
	def clear_error(self):
		self.error_position = None
		self.error_message = None

	def inspect(self, machine):
		if machine.is_inspectable:
			self.inspected_machine = machine
