################################################################
#
#   Assembly Line - Assembly robot
#
################################################################

from math import pi, sin, cos
from pygame import Rect
from albow.resource import get_image, get_font
from albow.image_array import get_image_array
from albow.controls import Label, Row, Column, Grid, IntEditor
from imath import iround
from machine import Machine, MachineNotConfiguredError
from game import game
from controls import MachineEditor
from part import Part

#------------------------------------------------------------------------

class AssemblyRobotEditor(MachineEditor):

	def create_contents(self, machine):
		return Row([
			Label("X Offset"), OffsetField('xoffset', machine),
			Label("Y Offset"), OffsetField('yoffset', machine),
		])

def OffsetField(attr, machine):
	return IntEditor(min = -59, max = 59, attribute = attr, model = machine)

#------------------------------------------------------------------------

class AssemblyRobot(Machine):
	#  xoffset   int
	#  yoffset   int
	
	title = "Assembly Robot"
	cost = 1000
	layer = 1
	icon = get_image("assembly_robot_icon.png")
	base_image = get_image("assembly_robot_base.png")
	arm_images = get_image_array("assembly_robot_arm.png", 9, border = 2)
	editor_class = AssemblyRobotEditor
	max_arm_phase = 8

	local_bounding_rect = Rect(-64, -96, 128, 192)
	upper_handling_rect = Rect(-2, -96, 4, 64)
	lower_handling_rect = Rect(-3, 32, 6, 64)
	local_delivery_point = (0, 64)
	arm_length = 64

	local_covered_rects = [
		[Rect(-32, -32, 64, 64)],
		[Rect(-32, -96, 64, 192), Rect(-64, -64, 32, 128), Rect(32, -64, 32, 128)]
	]
	
	xoffset = 0
	yoffset = 0
	arm_phase = 0
	carrying_part = None
	holding_part = None
	last_assembled_part = None
	
	def removed_from_factory(self):
		self.carrying_part = None
		self.holding_part = None
		self.last_assembled_part = None
	
	def draw_local(self, surf, t):
		#print "AssemblyRobot.draw_local: arm_phase =", self.arm_phase ###
		t.draw_image(surf, [self.base_image], (-32, -32))
		t.draw_image(surf, [self.arm_images[self.arm_phase]], (-68, -68))
	
	def check_ready(self):
		if self.xoffset is None or self.yoffset is None:
			raise MachineNotConfiguredError

	def handle_part(self, factory, part):
		#print "AssemblyRobot: Considering", part ###
		if self.arm_phase == 0 and not self.carrying_part:
			#print "...checking upper handling rect" ###
			if self.part_in_rect(part, self.upper_handling_rect):
				#print "...inside" ###
				self.pick_up_part(factory, part)
				return
		if not self.holding_part and part is not self.last_assembled_part:
			#print "...checking lower handling rect" ###
			if self.part_in_rect(part, self.lower_handling_rect):
				#print "...inside" ###
				self.hold_part(factory, part)
	
	def pick_up_part(self, factory, part):
		self.carrying_part = part
		part.handler = self
	
#	def drop_part(self, factory):
#		part = self.carrying_part
#		if part:
#			part.handler = None
#			self.carrying_part = None
	
	def hold_part(self, factory, part):
		self.holding_part = part
		part.handler = self
	
	def begin_frame(self, factory):
		if self.carrying_part:
			if self.arm_phase < self.max_arm_phase:
				self.arm_phase += 1
			else:
				if self.holding_part:
					self.perform_assembly(factory)
		else:
			if self.arm_phase > 0:
				self.arm_phase -= 1
		self.position_carried_part(factory)
	
	def position_carried_part(self, factory):
		part = self.carrying_part
		if part:
			a = pi * self.arm_phase / self.max_arm_phase
			r = self.arm_length
			x = iround(-r * sin(a))
			y = iround(-r * cos(a))
			p = self.global_point((x, y))
			factory.move_part_to(part, p)
	
	def perform_assembly(self, factory):
		part1 = self.holding_part
		part2 = self.carrying_part
		rect1 = part1.global_rect()
		rect2 = part2.pixels.get_rect()
		rect2.left = rect1.left + self.xoffset
		rect2.top = rect1.top + self.yoffset
		bounds = rect1.union(rect2)
		dx, dy = bounds.topleft
		rect1.move_ip(-dx, -dy)
		rect2.move_ip(-dx, -dy)
		bounds.move_ip(-dx, -dy)
		rmax = Rect(0, 0, 60, 60)
		rect1 = rect1.clip(rmax)
		rect2 = rect2.clip(rmax)
		bounds = bounds.clip(rmax)
		part3 = Part(bounds.size)
		pixels1, alpha1 = part1.arrays()
		pixels2, alpha2 = part2.arrays()
		pixels3, alpha3 = part3.arrays()
		xr1 = slice(rect1.left, rect1.right)
		xr2 = slice(rect2.left, rect2.right)
		yr1 = slice(rect1.top, rect1.bottom)
		yr2 = slice(rect2.top, rect2.bottom)
		pixels3[xr1, yr1] = pixels1
		pixels3[xr2, yr2] += pixels2
		alpha3[xr1, yr1] = alpha1
		alpha3[xr2, yr2] |= alpha2
		self.holding_part = None
		self.carrying_part = None
		self.last_assembled_part = part3
		part3.tag = 'assembly robot product'
		p = self.global_point(self.local_delivery_point)
		#print "AssemblyRobot: Finishing assembly:" ###
		#print "Removing", part1 ###
		#print "Removing", part2 ###
		#print "Adding", part3, "at", p ###
		factory.remove_part(part1)
		factory.remove_part(part2)
		factory.add_part(part3, p)

