################################################################
#
#   Assembly Line - Input Hopper
#
################################################################

from pygame import Rect
from albow.widget import Widget
from albow.controls import Label, Button, Row, Column, Grid, IntEditor, \
	ValueView
from albow.dialogs import Dialog, alert
from albow.resource import get_image, get_font
from machine import Machine, MachineNotConfiguredError
from settings import game_minutes_to_frames
from part import create_material
from material import MaterialType, available_sheet_sizes
from controls import TextMultichoice
from game import game

class InputHopperEditor(Dialog):

	title_font = get_font(18, "VeraBd.ttf")

	def __init__(self, machine):
		Dialog.__init__(self)
		self.machine = machine
		title = Label("Configure " + machine.title, font = self.title_font)
		rows = [
			[Label("Material"), MaterialField(machine.material)],
			#[Label("Sheet Size"), SheetSizeField(machine)],
			[Label("Sheet Length"), SheetSizeField('sheet_width', machine)],
			[Label("Sheet Width"), SheetSizeField('sheet_height', machine)],
			[Label("Supply Rate"), SupplyRateField(machine)],
			[Label("Material Price"), PixelCostView(machine),],
			[Label("Stock Cost"), SheetCostView(machine),],
		]
		grid = Grid(rows)
		title_and_grid = Column([title, grid], spacing = 16, align = 'c')
		ok_button = Button("OK", self.ok)
		contents = Column([title_and_grid, ok_button], align = 'r')
		contents.topleft = (16, 16)
		self.add(contents)
		self.shrink_wrap()
	
	def ok(self):
		self.dismiss(True)

#------------------------------------------------------------------------

def MaterialField(model):
	r = MaterialComponentField("R", "r", model)
	g = MaterialComponentField("G", "g", model)
	b = MaterialComponentField("B", "b", model)
	result = Row([r, g, b])
	r.focus()
	return result

def MaterialComponentField(title, attr, model):
	f = IntEditor(min = 0, max = 15, empty = None, attribute = attr, model = model)
	return Row([Label(title), f])

#def SheetSizeField(machine):
#	w = IntEditor(min = 1, max = 60, empty = None, attribute = "sheet_width", model = machine)
#	h = IntEditor(min = 1, max = 60, empty = None, attribute = "sheet_height", model = machine)
#	return Row([w, Label("x"), h])

def SheetSizeField(attribute, machine):
	return TextMultichoice(available_sheet_sizes, margin = 2, border_width = 1,
		attribute = attribute, model = machine)

def SupplyRateField(machine):
	f = IntEditor(min = 0, max = 60, empty = None, attribute = "supply_rate", model = machine)
	return Row([f, Label("sheets per hour")])

def PixelCostView(machine):
	return ValueView(format = "$%.2f per kilopixel",
		width = "$999999 per kilopixel",
		attribute = 'pixel_cost', model = machine)

def SheetCostView(machine):
	return ValueView(format = "$%.2f per sheet",
		width = "$999999 per sheet",
		attribute = 'sheet_cost', model = machine)

#------------------------------------------------------------------------

class InputHopper(Machine):
	#  material       MaterialType
	#  _sheet_width    int
	#  _sheet_height   int
	#  supply_rate    int
	#  order_quantity int
	#  timeout        int

	title = "Supply Hopper"
	icon = get_image("input_hopper_icon.png")
	on_left = True
	layer = 1
	image = get_image("input_hopper.png")
	cost = 1000
	editor_class = InputHopperEditor
	
	orgx, orgy = -32, -64
	
	local_bounding_rect = Rect(orgx, orgy, 64, 96)

	local_covered_rects = [
		[Rect(orgx, orgy, 64, 32)],
		[Rect(orgx, orgy, 64, 96)]
	]
	
	local_delivery_point = (orgx + 32, orgy + 64)
	
	_sheet_width = None
	_sheet_height = None
	supply_rate = None
	order_quantity = 20
	timeout = 0
	
	def __init__(self):
		self.material = MaterialType()
	
	def __setstate__(self, d):
		if 'sheet_width' in d:
			d['_sheet_width'] = d.pop('sheet_width')
		if 'sheet_height' in d:
			d['_sheet_height'] = d.pop('sheet_height')
		self.__dict__.update(d)
		if 'material' not in d:
			self.material = MaterialType()
	
	def get_pixel_cost(self):
		return self.material.cost
	
	pixel_cost = property(get_pixel_cost)
	
	def get_sheet_width(self):
		return self._sheet_width
	
	def get_sheet_height(self):
		return self._sheet_height
	
	def set_sheet_width(self, x):
		if x < 10:
			self._sheet_height = max(10, self._sheet_height)
		self._sheet_width = x
	
	def set_sheet_height(self, x):
		if x < 10:
			self._sheet_width = max(10, self._sheet_width)
		self._sheet_height = x
	
	sheet_width = property(get_sheet_width, set_sheet_width)
	sheet_height = property(get_sheet_height, set_sheet_height)
	
	def get_sheet_size(self):
		return (self.sheet_width, self.sheet_height)
	
	sheet_size = property(get_sheet_size)
	
	def get_sheet_cost(self):
		return self.material.sheet_cost(self.sheet_size)
	
	sheet_cost = property(get_sheet_cost)
	
	def check_ready(self):
		self.material.check_ready()
		if self.sheet_width is None or self.sheet_height is None \
			or self.supply_rate is None or self.order_quantity is None:
				raise MachineNotConfiguredError
	
	def draw_local(self, surf, t):
		org = self.local_topleft()
		t.draw_image(surf, [self.image], org)

	def begin_frame(self, factory):
		if self.timeout > 0:
			self.timeout -= 1
		else:
			self.deliver_part(factory)
			game_minutes = 60.0 / self.supply_rate
			self.timeout = game_minutes_to_frames(game_minutes)

	def deliver_part(self, factory):
		#print "InputHopper.deliver_part" ###
		company = game.company
		material = self.material
		size = self.sheet_size
		ord_qty = self.order_quantity
		part = company.get_material_sheet_from_stock(material, size, ord_qty)
		if part:
			part.tag = 'raw material'
			#print "InputHopper: Delivering", part ###
			p = self.global_point(self.local_delivery_point)
			factory.add_part(part, p)
		else:
			factory.stop()
			alert("Factory stopped because a supply hopper is out of stock.")
