################################################################
#
#   Assembly Line - Punch
#
################################################################

from math import pi
from pygame import Rect
from albow.widget import Widget
from albow.controls import Label, Button, Row, Column, Grid, IntEditor
from albow.dialogs import Dialog
from albow.resource import get_image, get_font
from albow.image_array import get_image_array
from machine import Machine, MachineNotConfiguredError
from controls import ImageMultichoice
from part import Part
from game import game

class PunchTool(object):
	#  shape   'rect' or 'circle'
	#  _size   (width, height)
	#  purchased  boolean
	
	title = "Punching Tool"
	base_cost = 50
	area_cost = 0.5
	perimeter_cost = 1
	
	purchased = False
	_size = (None, None)
	
	def __init__(self, shape, size):
		self.shape = shape
		self.size = size
	
	def get_size(self):
		if self.shape == 'circle':
			d = self._size[0]
			return (d, d)
		else:
			return self._size
		
	def set_size(self, x):
		self._size = x
	
	size = property(get_size, set_size)
	
	def calculate_cost(self):
		shape = self.shape
		width, height = self.size
		if shape == 'rect':
			area = width * height
			perimeter = 2 * (width + height)
		elif shape == 'circle':
			r = width
			area = pi * r * r
			perimeter = 2 * pi * r
		else:
			raise ValueError("Invalid tool shape")
		return int(self.base_cost +
			self.area_cost * area +
			self.perimeter_cost * perimeter)
	
	cost = property(calculate_cost) 
	
	def check_ready(self):
		if not self.shape or not self.size:
			raise MachineNotConfiguredError

#------------------------------------------------------------------------

shape_images = get_image_array("punch_shape_choices.png", 2, border = 2)
edge_images = get_image_array("punch_edge_choices.png", (2, 2), border = 2)

class PunchEditor(Dialog):

	title_font = get_font(18, "VeraBd.ttf")
	
	def __init__(self, machine):
		Widget.__init__(self)
		self.machine = machine
		title = Label("Configure " + machine.title, font = self.title_font)
		rows = [
			[Label("Tool Shape"), ToolShapeField(machine)],
			[ToolSizeLabel(machine), ToolSizeField(machine)],
			[Label("Offset"), OffsetField(machine)],
			[None, EdgeField(machine)],
		]
		grid = Grid(rows)
		title_and_grid = Column([title, grid], spacing = 16, align = 'c')
		ok_button = Button("OK", self.ok)
		contents = Column([title_and_grid, ok_button], align = 'r')
		contents.topleft = (16, 16)
		self.add(contents)
		self.shrink_wrap()
			
	def ok(self):
		self.dismiss(True)

def ToolShapeField(machine):
	return ImageMultichoice(images = shape_images,
		values = ['rect', 'circle'], attribute = 'tool_shape', model = machine)

def ShapeDependentField(machine, f1, f2):
	def is_rect():
		return machine.tool_shape == 'rect'
	def is_circle():
		return machine.tool_shape == 'circle'
	f1.get_visible = is_rect
	f2.get_visible = is_circle
	widget = Widget()
	widget.add([f1, f2])
	widget.shrink_wrap()
	return widget

def ToolSizeLabel(machine):
	lbl1 = Label("Tool Size")
	lbl2 = Label("Tool Diameter")
	return ShapeDependentField(machine, lbl1, lbl2)

def ToolSizeField(machine):
	f1 = ToolWidthHeightField(machine)
	f2 = ToolDiameterField(machine)
	return ShapeDependentField(machine, f1, f2)

def DimensionField(min, max, attr, machine):
	return IntEditor(min = min, max = max, empty = None,
		attribute = attr, model = machine)

def ToolWidthHeightField(machine):
	wf = DimensionField(1, 60, 'tool_width', machine)
	hf = DimensionField(1, 60, 'tool_height', machine)
	return Row([wf, Label("x"), hf])

def ToolDiameterField(machine):
	return DimensionField(1, 60, 'tool_width', machine)

def OffsetField(machine):
	return DimensionField(-30, 30, 'offset', machine)

class EdgeImages(object):

	indexes = {'rect': 0, 'circle': 1}

	def __init__(self, machine):
		self.machine = machine
	
	def __len__(self):
		return 2
	
	def __getitem__(self, i):
		j = self.indexes[self.machine.tool_shape]
		return edge_images[j, i]

def EdgeField(machine):
	return ImageMultichoice(images = EdgeImages(machine),
		values = ['left', 'right'], attribute = 'edge', model = machine)

#------------------------------------------------------------------------

class Punch(Machine):
	#  tool             PunchTool
	#  offset           int
	#  edge             'left' or 'right'
	#  delivery_side    'left' or 'right'
	#  state            'idle', 'extending', 'retracting'
	#  timeout          int
	#  handling_part    Part
	#  collected_parts  [Part]
	#  produced_parts   [Part]
	
	title = "Punch"
	icon = get_image("punch_icon.png")
	layer = 1
	image = get_image("punch.png")
	cost = 2000
	editor_class = PunchEditor
	cycle_frames = 5
	resting_frames = 5
	position_tolerance = 4
	max_extension = 10
	
	local_bounding_rect = Rect(-32, -96, 64, 128)

	local_covered_rects = [
		[Rect(-32, -64, 64, 32)],
		[local_bounding_rect]
	]

	local_input_rect = Rect(-32, -32, 64, 64)
	local_delivery_point = (0, -96)
	
	tool_image_array = get_image_array("punch_tools.png", 2)
	tool_images = {
		'circle': tool_image_array[0],
		'rect': tool_image_array[1],
	}
	
	tool = None
	offset = 0
	edge = 'left'
	state = 'idle'
	timeout = 0
	handling_part = None
	
	def __init__(self):
		self.collateral_parts = []
		self.produced_parts = []
		#<<<
		self.tool = PunchTool('circle', (25, 25))
		self.delivery_side = 'right'
		#>>>
	
	def __setstate__(self, d):
		if 'produced_parts' not in d:
			d['produced_parts'] = []
		self.__dict__.update(d)
	
	def removed_from_factory(self):
		self.release_part()
	
	def check_ready(self):
		if not self.tool:
			raise MachineNotConfiguredError
		self.tool.check_ready()
	
	def find_unpurchased_components(self, result):
		if not self.tool.purchased:
			result.append(self.tool)
	
	def get_tool_shape(self):
		if self.tool:
			return self.tool.shape
	
	def get_tool_size(self):
		if self.tool:
			return self.tool.size
	
	def get_tool_width(self):
		size = self.tool_size
		if size:
			return size[0]
		
	def get_tool_height(self):
		size = self.tool_size
		if size:
			return size[1]
		
	def set_tool_shape(self, x):
		self.configure_tool(x, self.tool.size)
	
	def set_tool_size(self, x):
		self.configure_tool(self.tool.shape, x)
	
	def set_tool_width(self, x):
		self.tool_size = (x, self.tool_height)
	
	def set_tool_height(self, x):
		self.tool_size = (self.tool_width, x)
	
	tool_shape = property(get_tool_shape, set_tool_shape)
	tool_size = property(get_tool_size, set_tool_size)
	tool_width = property(get_tool_width, set_tool_width)
	tool_height = property(get_tool_height, set_tool_height)

	def configure_tool(self, shape, size):
		def match(tool):
			return tool and tool.shape == shape and tool.size == size
		company = game.company
		if not match(self.tool):
			if self.tool:
				if self.tool.purchased:
					company.put_machine_into_storage(self.tool)
					self.tool = company.get_machine_from_storage(PunchTool, match)
				else:
					self.tool = None
			if not self.tool:
				self.tool = PunchTool(shape, size)

	def begin_frame(self, factory):
		if self.timeout > 0:
			self.timeout -= 1
		else:
			self.change_state(factory)
		del self.collateral_parts[:]
	
	def change_state(self, factory):
		state = self.state
		if state == 'idle':
			pass
		elif state == 'extending':
			self.process_parts(factory)
			self.state = 'retracting'
			self.timeout = self.cycle_frames
		elif state == 'retracting':
			self.deliver_parts(factory)
			self.release_part()
			self.state = 'resting'
			self.timeout = self.resting_frames
		else:
			self.state = 'idle'
		#print "Punch.change_state:", state, "-->", self.state
	
	def handle_part(self, factory, part):
		#print "Punch: Deciding whether to handle part", part ###
		result = False
		if self.state == 'idle' and not self.handling_part and self.part_in_input_rect(part):
			d = self.part_misalignment(part)
			#print "Punch: misalignment =", d ###
			if abs(d) <= self.position_tolerance:
				#print "Punch: Handling part", part ###
				factory.move_part(part, (-d, 0))
				self.handling_part = part
				part.handler = self
				self.state = 'extending'
				self.timeout = self.cycle_frames
				result = True
		if not part.handler:
			#print "Punch: Adding collateral part", part ###
			self.collateral_parts.append(part)
		#print "Punch: Decided whether to handle part:", result ###
		return result

	def part_in_input_rect(self, part):
		p = part.position
		r = self.global_rect(self.local_input_rect)
		return r.collidepoint(p)
	
	def part_misalignment(self, part):
		x, y = self.global_origin()
		r = part.global_rect()
		edge = self.edge
		w = self.tool.size[0] // 2
		if edge == 'left':
			d = r.left - (x - w - self.offset)
		else:
			d = r.right - (x + w + self.offset)
		return d

	def process_parts(self, factory):
		part = self.handling_part
		if part:
			#print "Punch: Processing handled part", part ###
			self.process_part(factory, part)
		for part in self.collateral_parts:
			#print "Punch: Processing collateral part", part ###
			self.process_part(factory, part)
	
	def release_part(self):
		part = self.handling_part
		if part:
			part.handler = None
			self.handling_part = None

	def process_part(self, factory, input_part):
		tool_part = get_tool_part(self.tool)
		tool_pixels = tool_part.pixels
		tool_rect = tool_part.global_rect()
		tool_rect.center = self.global_origin()
		
		input_pixels = input_part.pixels
		input_rect = input_part.global_rect()
		
		off = self.offset
		if self.edge == 'left':
			tool_rect.left = input_rect.left + off
		else:
			tool_rect.right = input_rect.right - off
		
		output_rect = input_rect.clip(tool_rect)
		ox, oy, ow, oh = output_rect
		if not (ow or oh):
			return
		output_part = Part(output_rect.size)
		output_pixels = output_part.pixels
		
		ix, iy = input_rect.topleft
		tx, ty = tool_rect.topleft
		hrange = xrange(ox, ox + ow)
		vrange = xrange(oy, oy + oh)
		e = (0, 0, 0, 0)
		empty = True
		#print "Punch.process_part:" ###
		#print "...input position =", ix, iy, "size =", input_rect.size ###
		#print "...tool position =", tx, ty, "size =", tool_rect.size ###
		#print "...output position =", ox, oy, "size =", output_rect.size ###
		for y in vrange:
			for x in hrange:
				a = tool_pixels.get_at((x - tx, y - ty))[3]
				if a:
					c = input_pixels.get_at((x - ix, y - iy))
					output_pixels.set_at((x - ox, y - oy), c)
					input_pixels.set_at((x - ix, y - iy), e)
					if c[3]:
						empty = False
		
		if not empty:
			input_part.tag = 'punch waste'
			#print "Punch: Produced", input_part ###
			self.produced_parts.append(output_part)
		
	def deliver_parts(self, factory):
		for part in self.produced_parts:
			self.deliver_part(factory, part)
		del self.produced_parts[:]

	def deliver_part(self, factory, part):
		part.tag = 'punch product'
		#print "Punch: Delivering", part ###
		p = self.global_point(self.local_delivery_point)
		factory.add_part(part, p)
	
	def draw_local(self, surf, t):
		state = self.state
		if state == 'extending' or state == 'retracting':
			phase = self.max_extension * self.timeout // self.cycle_frames
			if state == 'retracting':
				phase = self.max_extension - phase
			image = self.tool_images[self.tool.shape]
			p = (0, -phase)
			t.draw_image(surf, [image], p, ref = 'midbottom')
		p = (-32, -84)
		t.draw_image(surf, [self.image], p)

#------------------------------------------------------------------------

tool_parts = {}

def get_tool_part(tool):
	key = (tool.shape, tool.size)
	part = tool_parts.get(key)
	if not part:
		part = create_tool_part(tool)
		tool_parts[key] = part
	return part

def create_tool_part(tool):
	part = Part(tool.size)
	shape = tool.shape
	width, height = tool.size
	c = (0, 0, 0, 255)
	if shape == 'rect':
		part.pixels.fill(c)
	elif shape == 'circle':
		r = 0.5 * width
		part.fill_float_circle(c, (r, r), r)
	else:
		raise ValueError("Invalid tool shape")
	return part
