################################################################
#
#   Assembly Line - Material type
#
################################################################

from albow.resource import get_font
from imath import iceil
from machine import MachineNotConfiguredError

pixel_cost_matrix = [
	# k   w    y    m    c    r    g    b
	  1,  4,  10,  12,  14,  10,  300,  1000
]
pixel_cost_scale_factor = 0.1

base_cost = 1
area_factor = 1.0
perimeter_factor = 0.02
sheet_cost_scale_factor = 1.0

available_sheet_sizes = [1, 2, 5, 10, 20, 30, 40, 60]

formula_font = get_font(15, "Vera.ttf")
formula_subscript_font = get_font(10, "Vera.ttf")

#------------------------------------------------------------------------

class MaterialType:

	r = None
	g = None
	b = None
	
	def check_ready(self):
		if self.r is None or self.g is None or self.b is None:
			raise MachineNotConfiguredError
	
	def rgba(self):
		r = self.r or 0
		g = self.g or 0
		b = self.b or 0
		return (r * 0x11, g * 0x11, b * 0x11, 0xff)
	
	def get_cost(self):
		return pixel_cost(self.rgba())
	
	cost = property(get_cost)

	def sheet_cost(self, (width, height)):
		pc = pixel_cost(self.rgba())
		width = width or 0
		height = height or 0
		area = width * height * 0.001
		perimeter = 2 * (width + height)
		total = base_cost + pc * area_factor * area + perimeter_factor * perimeter
		#print "Material.sheet_cost: area", area, "perimeter", perimeter, "total", total ###
		#print base_cost, "+", pc * area_factor * area, "+", perimeter_factor * perimeter, "=", total ###
		return sheet_cost_scale_factor * total
	
	def formula(self):
		return text_formula(self.rgba())
	
#------------------------------------------------------------------------

def pixel_cost(rgba):
	r, g, b, a = rgba
	if not a:
		return 0
	r >>= 4; g >>= 4; b >>= 4; a >>= 4
	c = m = y = w = 0
	k = 45 - r - g - b
	w = min(r, g, b)
	r -= w; g -= w; b -= w
	c = min(g, b)
	m = min(r, b)
	y = min(r, g)
	if c: g -= c; b -= c
	if m: r -= m; b -= m
	if y: r -= y; g -= y
	assert r >= 0 and g >= 0 and b >= 0 and \
		c >= 0 and m >= 0 and y >= 0 and w >= 0
	total = 0
	for x, p in zip(pixel_cost_matrix, (k, w, y, m, c, r, g, b)):
		total += x * p
	return pixel_cost_scale_factor * total

def text_formula(color):
	if color[:3] == (0, 0, 0):
		return "K"
	else:
		items = []
		for c, n in zip("RGB", color):
			n >>= 4
			if n:
				items.append("%s%s" % (c, n))
		return ",".join(items)

def draw_formula(surf, color, topleft, fg, empty = None):
	x, y = topleft
	a = formula_font.get_ascent()
	if not color[3]:
		if empty:
			buf = formula_font.render(empty, True, fg)
			surf.blit(buf, (x, y))
	else:
		if color[:3] == (0, 0, 0):
			buf = formula_font.render("K", True, fg)
			surf.blit(buf, (x, y))
		else:
			for c, n in zip("RGB", color):
				n >>= 4
				if n:
					buf = formula_font.render(c, True, fg)
					r1 = buf.get_rect()
					r1.topleft = x, y
					surf.blit(buf, r1)
					buf = formula_subscript_font.render(str(n), True, fg)
					r2 = buf.get_rect()
					r2.midleft = r1.right, y + a
					surf.blit(buf, r2)
					x = r2.right
