################################################################
#
#   Assembly Line - Parts
#
################################################################

from __future__ import division

from math import floor, ceil
from pygame import Surface, SRCALPHA
from pygame.surfarray import pixels3d, pixels_alpha, make_surface
from Numeric import fromstring, reshape, zeros, asarray, add
from imath import iround, ifloor, iceil
from material import pixel_cost

class Part(object):
	#  pixels     Surface
	#  position   (int, int)
	#  handler    Machine
	
	handler = None
	tag = None
	
	def __init__(self, size):
		self.pixels = Surface(size, SRCALPHA, 32)
		self.position = (0, 0)
	
	def __str__(self):
		if self.tag:
			h = self.handler
			if h:
				hn = " handler " + h.__class__.__name__
			else:
				hn = ""
			return "<Part %r%s at 0x%x>" % (self.tag, hn, id(self))
		else:
			return repr(self)
	
	def __getstate__(self):
		pixels = self.pixels
		rgb = pixels3d(pixels)
		alpha = pixels_alpha(pixels)
		d = self.__dict__.copy()
		d['size'] = rgb.shape[:2]
		d['rgb'] = rgb.tostring()
		d['alpha'] = alpha.tostring()
		del d['pixels']
		return d
	
	def __setstate__(self, d):
		if 'shape' in d:
			self.old__setstate__(d)
			return
		size = d.pop('size')
		width, height = size
		rgb_data = reshape(fromstring(d.pop('rgb'), 'b'), (width, height, 3))
		alpha_data = reshape(fromstring(d.pop('alpha'), 'b'), (width, height))
		pixels = Surface(size, SRCALPHA, 32)
		rgb = pixels3d(pixels)
		rgb[:,:,:] = rgb_data
		alpha = pixels_alpha(pixels)
		alpha[:,:] = alpha_data
		d['size'] = size
		d['pixels'] = pixels
		self.__dict__.update(d)
		
	def old__setstate__(self, d):
		ary = fromstring(d.pop('data'), d.pop('typecode'))
		shape = d.pop('shape')
		print "Part.__setstate__: shape =", shape ###
		if len(shape) <> 3:
			raise EnvironmentError("Part data has wrong shape")
		ary = reshape(ary, shape)
		pixels = make_surface(ary)
		if shape[2] <> 4:
			print "Part.__setstate__: converting to alpha surface" ###
			pixels32 = Surface(pixels.get_size(), SRCALPHA, 32)
			pixels32.blit(pixels, (0, 0))
			pixels = pixels32
			print "...corner pixel =", pixels.get_at((0, 0)) ###
		d['pixels'] = pixels
		self.__dict__.update(d)
	
	def get_rect(self):
		return self.pixels.get_rect()
	
	def global_rect(self):
		r = self.pixels.get_rect()
		r.center = self.position
		return r
	
	def fill_float_rect(self, color, frect):
		l, t, r, b = map(iround, frect)
		rect = (l, t, r - l, b - t)
		self.pixels.fill(color, rect)
	
	def fill_float_circle(self, color, centre, radius):
		cx, cy = centre
		r = radius
		rsq = r * r
		pixels = self.pixels
		w, h = pixels.get_size()
		hrange = xrange(max(0, ifloor(cx - r)), min(w, iceil(cx + r)))
		vrange = xrange(max(0, ifloor(cy - r)), min(h, iceil(cy + r)))
		for y in vrange:
			for x in hrange:
				dx = x + 0.5 - cx
				dy = y + 0.5 - cy
				if dx * dx + dy * dy <= rsq:
					pixels.set_at((x, y), color)
	
	def arrays(self):
		pixels = self.pixels
		return pixels3d(pixels), pixels_alpha(pixels)
	
	def raw_material_cost(self):
		total = 0.0
		pix = self.pixels
		w, h = pix.get_size()
		for x in xrange(w):
			for y in xrange(h):
				c = pix.get_at((x, y))
				total += pixel_cost(c)
		return total * 0.001

#----------------------------------------------------------------------------

def create_material(material_type, size):
	part = Part(size)
	color = material_type.rgba()
	part.pixels.fill(color)
	return part

def compare_parts(part1, part2):
	pixels1 = part1.pixels
	pixels2 = part2.pixels
	rect1 = pixels1.get_rect()
	rect2 = pixels2.get_rect()
	rect2.center = rect1.center
	bounds = rect1.union(rect2)
	bounds.topleft = (0, 0)
	rect1.center = rect2.center = bounds.center
	bw, bh = bounds.size
	s = zeros((bw, bh, 4))
	t = s.typecode()
	x, y, w, h = rect1
	rgb = asarray(pixels3d(pixels1), t)
	alpha = asarray(pixels_alpha(pixels1), t)
	s[x:x+w, y:y+h, 0:3] += rgb >> 4
	s[x:x+w, y:y+h, 3] += alpha >> 4
	x, y, w, h = rect2
	rgb = asarray(pixels3d(pixels2), t)
	alpha = asarray(pixels_alpha(pixels2), t)
	s[x:x+w, y:y+h, 0:3] -= rgb >> 4
	s[x:x+w, y:y+h, 3] -= alpha >> 4
	r = add.reduce
	d = r(r(r(abs(s))))
	#d = r(r(r(s*s)))
	#print "compare_parts: d =", d ###
	return d / (bw * bh * 45)
